#!/usr/bin/env python3

'test that rumur-run can check a basic model'

import os
import platform
import subprocess as sp
import sys

if platform.system() == 'Linux' and 'TRAVIS' in os.environ:
  release = sp.check_output(['lsb_release', '--release', '--short'],
    universal_newlines=True).strip()
  if release == '14.04':
    if os.environ.get('CC') in ('gcc-7', 'gcc-8', 'gcc-9'):
      # XXX: When using GCC ≥7 on Ubuntu 14.04 (as we do in CI) and the compiler
      # is passed -march=native -mtune=native, it generates assembly sequences
      # that GAS from Binutils 2.24 (the default on this platform) is unable to
      # comprehend. It is possible to install Binutils 2.26 but in Travis it
      # seems flaky as to whether this works. To avoid false failures, just skip
      # this test in that configuration.
      print('this test is flaky on Ubuntu 14.04 with GCC 7+')
      sys.exit(125)


RUMUR_RUN = os.path.join(os.path.dirname(__file__), '../rumur/src/rumur-run')

MODEL = '''
var
  x: boolean;

startstate begin
  x := true;
end;

rule begin
  x := !x;
end;
'''

def main():
  p = sp.Popen(['python3', RUMUR_RUN], stdin=sp.PIPE)
  p.communicate(MODEL.encode('utf-8', 'replace'))
  return p.returncode

if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env python3

'''
Test that a compiled verifier does not depend on libatomic.

Uses of __atomic built-ins and C11 atomics can sometimes cause the compiler to
emit calls to libatomic instead of inline instructions. This is a problem
because we use these in the verifier to implement lock-free algorithms, while
the libatomic implementations take locks, defeating the purpose of using them.
This test checks that we end up with no libatomic calls in the compiled
verifier.
'''

import os
import platform
import re
import sys
import subprocess

# generate a checker for a simple model
model = 'var x: boolean; startstate begin x := false; end; rule begin x := !x; end;'
argv = ['rumur', '--output', '/dev/stdout']
p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
model_c, _ = p.communicate(model.encode('utf-8', 'replace'))
if p.returncode != 0:
  print('call to rumur failed')
  sys.exit(1)

# compile it to assembly
CC = os.environ.get('CC', 'cc')
argv = [CC, '-O3', '-std=c11', '-x', 'c', '-', '-S', '-o', '/dev/stdout'] \
  + sys.argv[1:]
p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
model_s, _ = p.communicate(model_c)
if p.returncode != 0:
  print('compilation failed')
  sys.exit(1)

# check for calls to libatomic functions
if re.search('__atomic_', model_s.decode('utf-8', 'replace')) is not None:
  print('libatomic calls in generated code were not optimised out')
  sys.exit(-1)

# pass
sys.exit(0)



#!/usr/bin/env python3

import argparse
import re
import sys

def main(args):

  # parse command line arguments
  parser = argparse.ArgumentParser(
    description='convert file contents to a C++ string')
  parser.add_argument('input', type=argparse.FileType('rb'), help='input file')
  parser.add_argument('output', type=argparse.FileType('wt'),
    help='output file')
  options = parser.parse_args(args[1:])

  array = re.sub(r'[^\w\d]', '_', options.input.name)
  size = '{}_len'.format(array)

  options.output.write(
    '#include <cstddef>\n'
    '\n'
    'extern const unsigned char {}[] = {{'.format(array))

  index = 0
  while True:

    c = options.input.read(1)
    if c == b'':
      break

    if index % 12 == 0:
      options.output.write('\n ')

    options.output.write(
      ' 0x{:02x},'.format(int.from_bytes(c, byteorder='little')))

    index += 1

  options.output.write(
    '\n'
    '}};\n'
    'extern const size_t {} = sizeof({}) / sizeof({}[0]);\n'
    .format(size, array, array))

  return 0

if __name__ == '__main__':
  sys.exit(main(sys.argv))



#!/usr/bin/env python3

'''
Version of lock-freedom.py for 32-bit compilation on x86-64
'''

import os
import platform
import sys
import subprocess

# this test is only relevant on x86-64
if platform.machine() not in ('amd64', 'x86_64'):
  print('only relevant for x86-64 machines')
  sys.exit(125)

# check that we have a multilib compiler capable of targeting i386
CC = os.environ.get('CC', 'cc')
argv = [CC, '-std=c11', '-m32', '-o', os.devnull, '-x', 'c', '-']
program = '#include <stdio.h>\n'           \
          'int main(void) {\n'             \
          '  printf("hello world\\n");\n'  \
          '  return 0;\n'                  \
          '}\n'
p = subprocess.Popen(argv, stdin=subprocess.PIPE, stderr=subprocess.DEVNULL)
p.communicate(program.encode('utf-8', 'replace'))
if p.returncode != 0:
  print('compiler cannot target 32-bit code')
  sys.exit(125)

# call the generic lock-freedom.py with some customisation
lock_freedom_py = os.path.join(os.path.dirname(__file__), 'lock-freedom.py')
sys.exit(subprocess.call(['python3', lock_freedom_py, '-m32']))



#!/usr/bin/env python3

'''
Generate contents of a version.cc.
'''

import os
import re
import subprocess as sp
import sys

def last_release():
  '''
  The version of the last tagged release of Rumur. This will be used as the
  version number if no Git information is available.
  '''
  with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
      '../../CHANGELOG.rst'), 'rt') as f:
    for line in f:
      m = re.match(r'(v\d{4}\.\d{2}\.\d{2})$', line)
      if m is not None:
        return m.group(1)

  raise Exception('version heading not found in changelog')

def has_git():
  '''
  Return True if we are in a Git repository and have Git.
  '''

  # Return False if we don't have Git.
  try:
    sp.check_call(['which', 'git'], stdout=sp.DEVNULL, stderr=sp.DEVNULL)
  except:
    return False

  # Return False if we have no Git repository information.
  if not os.path.exists(os.path.join(os.path.dirname(__file__),
      '..', '..', '.git')):
    return False

  return True

def get_tag():
  '''
  Find the version tag of the current Git commit, e.g. v2020.05.03, if it
  exists.
  '''
  try:
    tag = sp.check_output(['git', 'describe', '--tags'], stderr=sp.DEVNULL)
  except sp.CalledProcessError:
    tag = None

  if tag is not None:
    tag = tag.decode('utf-8', 'replace').strip()
    if re.match(r'v[\d\.]+$', tag) is None:
      # Not a version tag.
      tag = None

  return tag

def get_sha():
  '''
  Find the hash of the current Git commit.
  '''
  rev = sp.check_output(['git', 'rev-parse', '--verify', 'HEAD'])
  rev = rev.decode('utf-8', 'replace').strip()

  return rev

def is_dirty():
  '''
  Determine whether the current working directory has uncommitted changes.
  '''
  dirty = False

  ret = sp.call(['git', 'diff', '--exit-code'], stdout=sp.DEVNULL,
    stderr=sp.DEVNULL)
  dirty |= ret != 0

  ret = sp.call(['git', 'diff', '--cached', '--exit-code'], stdout=sp.DEVNULL,
    stderr=sp.DEVNULL)
  dirty |= ret != 0

  return dirty

def main(args):

  if len(args) != 2 or args[1] == '--help':
    sys.stderr.write('usage: {} file\n'.format(args[0]))
    sys.stderr.write(' write version information as a C++ source file\n')
    return -1

  # Get the contents of the old version file if it exists.
  old = None
  if os.path.exists(args[1]):
    with open(args[1], 'rt') as f:
      old = f.read()

  version = None

  # first, look for an environment variable that overrides other version sources
  version = os.environ.get('RUMUR_VERSION')

  # second, look for a version tag on the current commit
  if version is None and has_git():
    tag = get_tag()
    if tag is not None:
      version = '{}{}'.format(tag, ' (dirty)' if is_dirty() else '')

  # third, look for the commit hash as the version
  if version is None and has_git():
    rev = get_sha()
    assert rev is not None
    version = 'Git commit {}{}'.format(rev, ' (dirty)' if is_dirty() else '')

  # Finally, fall back to our known release version.
  if version is None:
    version = last_release()

  new = '#pragma once\n' \
        '\n' \
        'namespace rumur {{\n' \
        '\n' \
        'static constexpr const char *get_version() {{\n' \
        '  return "{}";\n' \
        '}}\n' \
        '\n' \
        '}}'.format(version)

  # If the version has changed, update the output. Otherwise we leave the old
  # contents -- and more importantly, the timestamp -- intact.
  if old != new:
    with open(args[1], 'wt') as f:
      f.write(new)

  return 0

if __name__ == '__main__':
  sys.exit(main(sys.argv))



#!/usr/bin/env python3

'''
x86-64 version of lock-freedom.py, which needs -mcx16
'''

import os
import platform
import sys
import subprocess

# this test is only relevant on x86-64
if platform.machine() not in ('amd64', 'x86_64'):
  print('only relevant for x86-64 machines')
  sys.exit(125)

# call the generic lock-freedom.py with some customisation
lock_freedom_py = os.path.join(os.path.dirname(__file__), 'lock-freedom.py')
sys.exit(subprocess.call(['python3', lock_freedom_py, '-mcx16']))



#!/usr/bin/env python3

'''
integration test suite

Despite using Python’s unittest, this is not a set of unit tests. The Python
module is simply a nice low-overhead testing framework.
'''

import codecs
import multiprocessing
import os
from pathlib import Path
import re
import subprocess as sp
import sys
import tempfile
import unittest

CPUS = multiprocessing.cpu_count()

VERIFIER_RNG = Path(__file__).resolve().parent / '../misc/verifier.rng'
MURPHI2XML_RNG = Path(__file__).resolve().parent / '../misc/murphi2xml.rng'

# test configuration variables, set during main
CONFIG = {}

def enc(s): return s.encode('utf-8', 'replace')
def dec(s): return s.decode('utf-8', 'replace')

def run(args, stdin = None):
  '''
  run a command and return its result
  '''
  if stdin is not None:
    stdin = enc(stdin)
  env = {k: v for k, v in os.environ.items()}
  env.update({k: str(v) for k, v in CONFIG.items()})
  p = sp.Popen([str(a) for a in args], stdout=sp.PIPE, stderr=sp.PIPE,
               stdin=sp.PIPE, env=env)
  stdout, stderr = p.communicate(stdin)
  return p.returncode, dec(stdout), dec(stderr)

def parse_test_options(src, debug = False, multithreaded = False, xml = False):
  '''
  extract test tweaks and directives from leading comments in a test input
  '''
  with open(str(src), 'rt', encoding='utf-8') as f:
    for line in f:
      # recognise '-- rumur_flags: …' etc lines
      m = re.match(r'\s*--\s*(?P<key>[a-zA-Z_]\w*)\s*:(?P<value>.*)$', line)
      if m is None:
        break
      yield m.group('key'), eval(m.group('value').strip())

class executable(unittest.TestCase):
  '''
  test cases involving running a custom executable file
  '''

  def _run(self, testcase):

    assert os.access(str(testcase), os.X_OK), 'non-executable test case ' \
      '{} attached to executable class'.format(testcase)

    ret, stdout, stderr = run([testcase])
    output = '{}{}'.format(stdout, stderr)
    if ret == 125:
      self.skipTest(output.strip())
    elif ret != 0:
      self.fail(output)

class murphi2c(unittest.TestCase):
  '''
  test cases for murphi2c
  '''

  def _run(self, testcase):

    tweaks = {k: v for k, v in parse_test_options(testcase)}

    # there is no C equivalent of isundefined, because an implicit assumption in
    # the C representation is that you do not rely on undefined values
    with open(str(testcase), 'rt', encoding='utf-8') as f:
      should_fail = re.search(r'\bisundefined\b', f.read()) is not None

    args = ['murphi2c', testcase]
    if CONFIG['HAS_VALGRIND']:
      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
        '--error-exitcode=42'] + args
    ret, stdout, stderr = run(args)
    if CONFIG['HAS_VALGRIND']:
      if ret == 42:
        self.fail('Memory leak:\n{}{}'.format(stdout, stderr))

    # if rumur was expected to reject this model, we allow murphi2c to fail
    if tweaks.get('rumur_exit_code', 0) == 0 and not should_fail and ret != 0:
      self.fail('Unexpected murphi2c exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if should_fail and ret == 0:
      self.fail('Unexpected murphi2c exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if ret != 0:
      return

    # ask the C compiler if this is valid
    args = [CONFIG['CC']] + CONFIG['C_FLAGS'] + ['-c', '-o', os.devnull, '-']
    ret, out, err = run(args, stdout)
    if ret != 0:
      self.fail('C compilation failed:\n{}{}\nProgram:\n{}'
                .format(out, err, stdout))

class murphi2cHeader(unittest.TestCase):
  '''
  test cases for murphi2c --header
  '''

  def _run(self, testcase):

    tweaks = {k: v for k, v in parse_test_options(testcase)}

    # there is no C equivalent of isundefined, because an implicit assumption in
    # the C representation is that you do not rely on undefined values
    with open(str(testcase), 'rt', encoding='utf-8') as f:
      should_fail = re.search(r'\bisundefined\b', f.read()) is not None

    args = ['murphi2c', '--header', testcase]
    if CONFIG['HAS_VALGRIND']:
      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
        '--error-exitcode=42'] + args
    ret, stdout, stderr = run(args)
    if CONFIG['HAS_VALGRIND']:
      if ret == 42:
        self.fail('Memory leak:\n{}{}'.format(stdout, stderr))

    # if rumur was expected to reject this model, we allow murphi2c to fail
    if tweaks.get('rumur_exit_code', 0) == 0 and not should_fail and ret != 0:
      self.fail('Unexpected murphi2c exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if should_fail and ret == 0:
      self.fail('Unexpected murphi2c exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if ret != 0:
      return

    with tempfile.TemporaryDirectory() as tmp:

      # write the header to a temporary file
      header = Path(tmp) / 'header.h'
      with open(str(header), 'wt', encoding='utf-8') as f:
        f.write(stdout)

      # ask the C compiler if the header is valid
      main_c = '#include "{}"\nint main(void) {{ return 0; }}\n'.format(header)
      args = [CONFIG['CC']] + CONFIG['C_FLAGS'] + ['-o', os.devnull, '-']
      ret, stdout, stderr = run(args, main_c)
      if ret != 0:
        self.fail('C compilation failed:\n{}{}'.format(stdout, stderr))

      # ask the C++ compiler if it is valid there too
      ret, stdout, stderr = run([CONFIG['CXX'], '-std=c++11', '-o', os.devnull,
        '-x', 'c++', '-', '-Werror=format', '-Werror=sign-compare',
        '-Werror=type-limits'], main_c)
      if ret != 0:
        self.fail('C++ compilation failed:\n{}{}'.format(stdout, stderr))

class murphi2uclid(unittest.TestCase):
  '''
  test cases for murphi2uclid
  '''

  def _run(self, testcase):

    tweaks = {k: v for k, v in parse_test_options(testcase)}

    # test cases for which murphi2uclid is expected to fail
    MURPHI2UCLID_FAIL = (
      # contains '<<' or '>>'
      'lsh-basic.m',
      'rsh-and.m',
      'rsh-basic.m',
      'smt-bv-lsh.m',
      'smt-bv-rsh.m',

      # contains '/'
      'division.m',
      'smt-bv-div.m',
      'smt-bv-div2.m',
      'smt-div.m',

      # contains '%'
      'put-string-injection.m',
      'smt-bv-mod.m',
      'smt-bv-mod2.m',
      'smt-mod.m',

      # contains alias statements
      'alias-and-field.m',
      'alias-in-bound.m',
      'alias-in-bound2.m',
      'alias-literal.m',
      'alias-of-alias-rule.m',
      'alias-of-alias-rule2.m',
      'alias-of-alias-stmt.m',
      'basic-aliasrule.m',
      'mixed-aliases.m',

      # 'clear' of a complex type
      'clear-complex.m',

      # contains 'cover'
      'cover-basic.m',
      'cover-basic2.m',
      'cover-miss.m',
      'cover-multiple.m',
      'cover-stmt.m',
      'cover-stmt-miss.m',
      'cover-trivial.m',
      'string-injection.m',

      # contains 'isundefined'
      'diff-trace-arrays.m',
      'isundefined-basic.m',
      'isundefined-decl.m',
      'isundefined-element.m',
      'isundefined-function.m',
      'for-variants.m',
      'scalarset-cex.m',
      'scalarset-schedules-off.m',
      'scalarset-schedules-off-2.m',

      # contains 'put'
      'for-step-0-dynamic.m',
      'put-stmt.m',
      'put-stmt2.m',
      'put-stmt3.m',
      'put-stmt4.m',
      'scalarset-put.m',

      # contains early return from a function/procedure/rule
      'return-from-rule.m',
      'return-from-ruleset.m',
      'return-from-startstate.m',

      # 'exists' or 'forall' with non-1 step
      'smt-bv-exists4.m',
      'smt-bv-forall4.m',
      'smt-exists4.m',
      'smt-forall4.m',

      # 'liveness' inside a 'ruleset'
      'liveness-in-ruleset.m',
      'liveness-in-ruleset2.m',
    )

    # test cases fo which Uclid5 is expected to fail
    UCLID_FAIL = (
      # contains a record field with the same name as a variable
      # https://github.com/uclid-org/uclid/issues/99
      'compare-record.m',
      'smt-array-of-record.m',
      'smt-record-bool-field.m',
      'smt-record-bool-field2.m',
      'smt-record-enum-field.m',
      'smt-record-enum-field2.m',
      'smt-record-of-array.m',
      'smt-record-range-field.m',
      'smt-record-range-field2.m',

      # recursive function calls
      'recursion2.m',
      'recursion4.m',

      # reference to a field of an array element
      '193.m',

      # function calls within expressions
      'differing-type-return.m',
      'differing-type-return3.m',
      'function-call-in-if.m',
      'function-in-guard.m',
      'multiple-parameters.m',
      'multiple-parameters2.m',
      'non-const-parameters.m',
      'recursion1.m',
      'recursion5.m',
      'reference-function-parameter.m',
      'reference-function-parameter2.m',
      'section-order4.m',
      'section-order5.m',
      'section-order10.m',
      'type-shadowing2.m',

      # modifies a mutable parameter within a function, which is not valid
      # within a Uclid5 procedure
      'reference-function-parameter3.m',
    )

    args = ['murphi2uclid', testcase]
    if CONFIG['HAS_VALGRIND']:
      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
        '--error-exitcode=42'] + args
    ret, stdout, stderr = run(args)
    if CONFIG['HAS_VALGRIND']:
      if ret == 42:
        self.fail('Memory leak:\n{}{}'.format(stdout, stderr))

    # if rumur was expected to reject this model, we allow murphi2uclid to fail
    should_fail = testcase.name in MURPHI2UCLID_FAIL
    could_fail = tweaks.get('rumur_exit_code', 0) != 0 or should_fail

    if not could_fail and ret != 0:
      self.fail('Unexpected murphi2uclid exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if should_fail and ret == 0:
      self.fail('Unexpected murphi2uclid exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if ret != 0:
      return

    # if we do not have Uclid5 available, skip the remainder of the test
    if not CONFIG['HAS_UCLID']:
      self.skipTest('uclid not available for validation')

    with tempfile.TemporaryDirectory() as tmp:

      # write the Uclid5 source to a temporary file
      src = Path(tmp) / 'source.ucl'
      with open(str(src), 'wt', encoding='utf-8') as f:
        f.write(stdout)

      # ask Uclid if the source is valid
      ret, stdout, stderr = run(['uclid', src])
      if testcase.name in UCLID_FAIL and ret == 0:
        self.fail('uclid unexpectedly succeeded:\n{}{}'.format(stdout, stderr))
      if testcase.name not in UCLID_FAIL and ret != 0:
        self.fail('uclid failed:\n{}{}'.format(stdout, stderr))

class murphi2xml(unittest.TestCase):
  '''
  test cases for murphi2xml
  '''

  def _run(self, testcase):

    tweaks = {k: v for k, v in parse_test_options(testcase)}

    args = ['murphi2xml', testcase]
    if CONFIG['HAS_VALGRIND']:
      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
        '--error-exitcode=42'] + args
    ret, stdout, stderr = run(args)
    if CONFIG['HAS_VALGRIND']:
      if ret == 42:
        self.fail('Memory leak:\n{}{}'.format(stdout, stderr))

    # if rumur was expected to reject this model, we allow murphi2xml to fail
    if tweaks.get('rumur_exit_code', 0) == 0 and ret != 0:
      self.fail('Unexpected murphi2xml exit status {}:\n{}{}'
                .format(ret, stdout, stderr))

    if ret != 0:
      return

    # murphi2xml will have written XML to its stdout
    xmlcontent = stdout

    # See if we have xmllint
    if not CONFIG['HAS_XMLLINT']:
      self.skipTest('xmllint not available for validation')

    # Validate the XML
    ret, stdout, stderr = run(['xmllint', '--relaxng', MURPHI2XML_RNG,
      '--noout', '-'], xmlcontent)
    if ret != 0:
      self.fail('Failed to validate:\n{}{}'.format(stdout, stderr))

class rumur(unittest.TestCase):
  '''
  test cases involving generating a checker and running it
  '''

  def _run_param(self, testcase, debug, optimised, multithreaded, xml):

    tweaks = {k: v for k, v in parse_test_options(testcase, debug,
      multithreaded, xml)}

    if tweaks.get('skip_reason') is not None:
      self.skipTest(tweaks['skip_reason'])

    # build up arguments to call rumur
    args = ['rumur', '--output', '/dev/stdout', testcase]
    if debug: args += ['--debug']
    if xml: args += ['--output-format', 'machine-readable']
    if multithreaded and CPUS == 1: args +=['--threads', '2']
    elif not multithreaded: args += ['--threads', '1']
    args += tweaks.get('rumur_flags', [])

    if CONFIG['HAS_VALGRIND']:
      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
        '--error-exitcode=42'] + args

    # call rumur
    ret, stdout, stderr = run(args)
    if CONFIG['HAS_VALGRIND']:
      if ret == 42:
        self.fail('Memory leak:\n{}{}'.format(stdout, stderr))
    if ret != tweaks.get('rumur_exit_code', 0):
      self.fail('Rumur failed:\n{}{}'.format(stdout, stderr))

    # if we expected to fail, we are done
    if ret != 0: return

    model_c = stdout

    with tempfile.TemporaryDirectory() as tmp:

      # build up arguments to call the C compiler
      model_bin = Path(tmp) / 'model.exe'
      args = [CONFIG['CC']] + CONFIG['C_FLAGS']
      if optimised: args += ['-O3']
      args += ['-o', model_bin, '-', '-lpthread']

      if CONFIG['NEEDS_LIBATOMIC']:
        args += ['-latomic']

      # call the C compiler
      ret, stdout, stderr = run(args, model_c)
      if ret != 0:
        self.fail('C compilation failed:\n{}{}'.format(stdout, stderr))

      # now run the model itself
      ret, stdout, stderr = run([model_bin])
      if ret != tweaks.get('checker_exit_code', 0):
        self.fail('Unexpected checker exit status {}:\n{}{}'
                  .format(ret, stdout, stderr))

    # if the test has a stdout expectation, check that now
    if tweaks.get('checker_output') is not None:
      if tweaks['checker_output'].search(stdout) is None:
        self.fail('Checker output did not match expectation regex:\n{}{}'
                  .format(stdout, stderr))

    # coarse grained check for whether the model contains a 'put' statement that
    # could screw up XML validation
    with open(str(testcase), 'rt', encoding='utf-8') as f:
      has_put = re.search(r'\bput\b', f.read()) is not None

    if xml and not has_put:

      model_xml = stdout

      if not CONFIG['HAS_XMLLINT']: self.skipTest('xmllint not available')

      # validate the XML
      args = ['xmllint', '--relaxng', VERIFIER_RNG, '--noout', '-']
      ret, stdout, stderr = run(args, model_xml)
      if ret != 0:
        self.fail('Failed to XML-validate machine reachable output:\n{}{}'
                  .format(stdout, stderr))

class rumurSingleThreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, False, False, False)

class rumurDebugSingleThreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, True, False, False, False)

class rumurOptimisedSingleThreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, True, False, False)

class rumurDebugOptimisedSingleThreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, True, True, False, False)

class rumurMultithreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, False, True, False)

class rumurDebugMultithreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, True, False, True, False)

class rumurOptimisedMultithreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, True, True, False)

class rumurDebugOptimisedMultithreaded(rumur):
  def _run(self, testcase):
    self._run_param(testcase, True, True, True, False)

class rumurSingleThreadedXML(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, False, False, True)

class rumurOptimisedSingleThreadedXML(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, True, False, True)

class rumurMultithreadedXML(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, False, True, True)

class rumurOptimisedMultithreadedXML(rumur):
  def _run(self, testcase):
    self._run_param(testcase, False, True, True, True)

def make_name(t):
  '''
  name mangle a path into a valid test case name
  '''
  safe_name = re.sub(r'[^a-zA-Z0-9]', '_', t.name)
  return 'test_{}'.format(safe_name)

def main():

  # setup stdout to make encoding errors non-fatal
  sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'replace')

  # parse configuration
  global CONFIG
  for p in sorted((Path(__file__).parent / 'config').iterdir()):

    # skip subdirectories
    if p.is_dir(): continue

    # skip non-executable files
    if not os.access(str(p), os.X_OK): continue

    CONFIG[p.name] = eval(sp.check_output([str(p)]))

  # find files in our directory
  root = Path(__file__).parent
  for p in sorted(root.iterdir()):

    # skip directories
    if p.is_dir(): continue

    # skip ourselves
    if os.path.samefile(str(p), __file__): continue

    name = make_name(p)

    # if this is executable, treat it as a test case
    if os.access(str(p), os.X_OK):
      assert not hasattr(executable, name), \
        'name collision involving executable.{}'.format(name)
      setattr(executable, name, lambda self, p=p: self._run(p))

    # if this is not a model, skip the remaining generic logic
    if p.suffix != '.m': continue

    for c in (rumurSingleThreaded,
              rumurDebugSingleThreaded,
              rumurOptimisedSingleThreaded,
              rumurDebugOptimisedSingleThreaded,
              rumurMultithreaded,
              rumurDebugMultithreaded,
              rumurOptimisedMultithreaded,
              rumurDebugOptimisedMultithreaded,
              rumurSingleThreadedXML,
              rumurOptimisedSingleThreadedXML,
              rumurMultithreadedXML,
              rumurOptimisedMultithreadedXML,
             ):
      assert not hasattr(c, name), \
        'name collision involving rumur.{}'.format(name)
      setattr(c, name, lambda self, p=p: self._run(p))

    assert not hasattr(murphi2c, name), \
      'name collision involving murphi2c.{}'.format(name)
    setattr(murphi2c, name, lambda self, p=p: self._run(p))

    assert not hasattr(murphi2cHeader, name), \
      'name collision involving murphi2cHeader.{}'.format(name)
    setattr(murphi2cHeader, name, lambda self, p=p: self._run(p))

    assert not hasattr(murphi2uclid, name), \
      'name collision involving murphi2uclid.{}'.format(name)
    setattr(murphi2uclid, name, lambda self, p=p: self._run(p))

    assert not hasattr(murphi2xml, name), \
      'name collision involving murphi2xml.{}'.format(name)
    setattr(murphi2xml, name, lambda self, p=p: self._run(p))

  unittest.main()

if __name__ == '__main__':
  main()



#!/usr/bin/env python3

import os
import pathlib
import re
import subprocess

# an arbitrary test model that has a liveness property
model = pathlib.Path(__file__).parent / 'liveness-miss1.m'
assert model.exists()

# confirm it contains the liveness property we expect
with open(str(model), 'rt', encoding='utf-8') as f:
  assert re.search(r'liveness "x is 10" x = 10', f.read())

# use the remove-liveness pass to remove the property
print('+ murphi2murphi --remove-liveness {}'.format(model))
transformed = subprocess.check_output(['murphi2murphi',
  '--remove-liveness', str(model)])
decoded = transformed.decode('utf-8', 'replace')

print('transformed model:\n{}'.format(decoded))

# now confirm the property is no longer present
assert re.search(r'liveness "x is 10" x = 10', decoded) is None

# the generated model also should be valid syntax for Rumur
print('+ rumur --output /dev/null <(transformed model)')
p = subprocess.Popen(['rumur', '--output', os.devnull], stdin=subprocess.PIPE)
p.communicate(transformed)
assert p.returncode == 0



#!/usr/bin/env python3

import os
import pathlib
import re
import subprocess

# pick an arbitrary test model that omits semi-colons
model = pathlib.Path(__file__).parent / 'assume-statement.m'
assert model.exists()

# use the explicit-semicolons pass to add semi-colons
print('+ murphi2murphi --explicit-semicolons {}'.format(model))
transformed = subprocess.check_output(['murphi2murphi',
  '--explicit-semicolons', str(model)])
decoded = transformed.decode('utf-8', 'replace')

print('transformed model:\n{}'.format(decoded))

# we should now find the variable definition and both rules in this model are
# semi-colon terminated
assert re.search(r'\bx: 0 \.\. 2;$', decoded, re.MULTILINE)
assert len(re.findall(r'^end;$', decoded, re.MULTILINE)) == 2

# the generated model also should be valid syntax for Rumur
print('+ rumur --output /dev/null <(transformed model)')
p = subprocess.Popen(['rumur', '--output', os.devnull], stdin=subprocess.PIPE)
p.communicate(transformed)
assert p.returncode == 0



#!/usr/bin/env python3

import os
import pathlib
import re
import subprocess

# model from the test directory involving some switch examples
model = pathlib.Path(__file__).parent / 'switch-stmt2.m'
assert model.exists()

# use the switch-to-if pass to remove the switch statements
print('+ murphi2murphi --switch-to-if {}'.format(model))
transformed = subprocess.check_output(['murphi2murphi',
  '--switch-to-if', str(model)])
decoded = transformed.decode('utf-8', 'replace')

print('transformed model:\n{}'.format(decoded))

# we should now be able to find some if statements in the model
assert re.search(r'\bif x = y then\b', decoded)
assert re.search(r'\belsif x = 5 then\b', decoded)

# the generated model also should be valid syntax for Rumur
print('+ rumur --output /dev/null <(transformed model)')
p = subprocess.Popen(['rumur', '--output', os.devnull], stdin=subprocess.PIPE)
p.communicate(transformed)
assert p.returncode == 0



#!/usr/bin/env python3

'''
test comment retrieval API
'''

import shutil
import subprocess
import sys

if not shutil.which('murphi-comment-ls'):
  print('murphi-comment-ls not found')
  sys.exit(125)


# a model without any comments
NONE = '''
var
  x: boolean;

startstate begin
  x := true;
end;

rule begin
  x := !x;
end;
'''

# a model with a single line comment
SINGLE = '''
var
  x: boolean;

startstate begin
  x := -- hello world
  true;
end;

rule begin
  x := !x;
end;
'''

# a model with a multiline comment
MULTILINE = '''
var
  x: boolean;

startstate begin
  x := /* hello world */ true;
end;

rule begin
  x := !x;
end;
'''

# a model with a single line comment terminated by EOF, not \n
SINGLE_EOF = '''
var
  x: boolean;

startstate begin
  x := true;
end;

rule begin
  x := !x;
end; -- hello world'''

# a model with a multiline comment terminated by EOF, not */
MULTILINE_EOF = '''
var
  x: boolean;

startstate begin
  x := true;
end;

rule begin
  x := !x;
end; /* hello world '''

# comments inside strings that should be ignored
STRING = '''
var
  x: boolean;

startstate "hello -- world" begin
  x := true;
end;

rule "hello /* world */" begin
  x := !x;
end;
'''

# a model with a mixture
MIX = '''
var
  x: boolean;

startstate begin
  x := -- hello world
  /* hello world */true;
end; -- hello /* hello world */
/* hello -- hello */
rule begin
  x := !x;
end;
'''

def process(input):
  proc = subprocess.Popen(['murphi-comment-ls'], stdin=subprocess.PIPE,
    stdout=subprocess.PIPE, universal_newlines=True)
  stdout, _ = proc.communicate(input)
  assert proc.returncode == 0
  return stdout

def main():

  assert process(NONE) == ''

  assert process(SINGLE) == '6.8-21:  hello world\n'
  assert process(MULTILINE) == '6.8-24:  hello world \n'

  assert process(SINGLE_EOF) == '11.6-19:  hello world\n'
  assert process(MULTILINE_EOF) == '11.6-20:  hello world \n'

  assert process(STRING) == ''

  assert process(MIX) == '6.8-21:  hello world\n' \
                         '7.3-19:  hello world \n' \
                         '8.6-31:  hello /* hello world */\n' \
                         '9.1-20:  hello -- hello \n'

  return 0

if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env python3

import os
import pathlib
import re
import subprocess

# model from the test directory involving a record comparison
model = pathlib.Path(__file__).parent / 'unicode-assignment.m'
assert model.exists()

# use the ASCII transformation to remove ≔
print('+ murphi2murphi --to-ascii {}'.format(model))
transformed = subprocess.check_output(['murphi2murphi',
  '--to-ascii', str(model)])
decoded = transformed.decode('utf-8', 'replace')

print('transformed model:\n{}'.format(decoded))

# the ≔ operator should have become :=
assert re.search(r'\bx := true\b', decoded)
assert re.search(r'\bx := !x\b', decoded)

# the generated model also should be valid syntax for Rumur
print('+ rumur --output /dev/null <(transformed model)')
p = subprocess.Popen(['rumur', '--output', os.devnull], stdin=subprocess.PIPE)
p.communicate(transformed)
assert p.returncode == 0



#!/usr/bin/env python3

'''
This is a variant of 193.m that tests that we descend into TypeExprIDs when
reordering record fields. It is hard to trigger a behaviour divergence that is
exposed during checking, so instead we check the --debug output of rumur.

See also https://github.com/Smattr/rumur/issues/193.
'''

import os
import subprocess as sp
import sys

# a model containing a TypeExprID, itself referring to a record whose fields
# will be reordered
MODEL = '''
type
  t1: scalarset(4);

  t2: record
    a: boolean;
    b: t1;
    c: boolean;
  end;

var x: t2;

startstate begin
end

rule begin
end
'''

def main():

  # run Rumur in debug mode
  output = sp.check_output(['rumur', '--output', os.devnull, '--debug'],
    stderr=sp.STDOUT, input=MODEL.encode('utf-8', 'replace'))
  output = output.decode('utf-8', 'replace')

  # we should see the fields be reordered once due to encountering the original
  # TypeDecl (t2) and once due to the TypeExprID (the type of x)
  assert output.count('sorted fields {a, b, c} -> {a, c, b}') == 2

  return 0

if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env python3

import os
import pathlib
import re
import subprocess

# model from the test directory involving a record comparison
model = pathlib.Path(__file__).parent / 'compare-record.m'
assert model.exists()

# use the complex comparison decomposition to explode the record comparison in
# this model
print('+ murphi2murphi --decompose-complex-comparisons {}'.format(model))
transformed = subprocess.check_output(['murphi2murphi',
  '--decompose-complex-comparisons', str(model)])
decoded = transformed.decode('utf-8', 'replace')

print('transformed model:\n{}'.format(decoded))

# the comparisons should have been decomposed into member-wise comparison
assert re.search(r'\bx\.x = y\.x\b', decoded)
assert re.search(r'\bx\.x != y\.x\b', decoded)

# the generated model also should be valid syntax for Rumur
print('+ rumur --output /dev/null <(transformed model)')
p = subprocess.Popen(['rumur', '--output', os.devnull], stdin=subprocess.PIPE)
p.communicate(transformed)
assert p.returncode == 0



#!/usr/bin/env python3

'basic test that rumur-run can execute successfully'

import os
import subprocess
import sys

RUMUR_RUN = os.path.join(os.path.dirname(__file__), '../rumur/src/rumur-run')

def main():
  subprocess.check_call(['python3', RUMUR_RUN, '--version'])

if __name__ == '__main__':
  sys.exit(main())
